#include "functionsBEM.hpp"

// Retrieves every usefull information from the inside domain for the post-processing. "domainNodeTags" and "domainNodeCoords"
// are vectors containing all the node tags and coordinates (respectivelly) of the domain. "domainNodeIndices" is a map
// associating the index of each node tag for accessing the "domainNodeCoords" vector. "domainEntityTags" is a vector containing
// the tag of each entity of the domain. "domainElementTypes" stores the types of the elements of the the domain. The first
// dimension of this vector represents the different entities and the second, the different element types. "domainElementTags"
// stores the tags of the elements of the domain. The first dimension of this vector represents the different entities, the
// second, the different element types and the third, the different element tags. The "domainElementNodeTags" map associates to
// each element tag a vector of node tags representing the nodes of that element. "domainPhysicalGroupTag" and
// "domainPhysicalGroupDim" are the dimension and tag (respectivelly) of the BEM domain of interrest.
void getModel(std::vector<std::size_t> &domainNodeTags,
              std::map<std::size_t, std::size_t> &domainNodeIndices,
              std::vector<double> &domainNodeCoords,
              std::vector<int> &domainEntityTags,
              std::vector<std::vector<int>> &domainElementTypes,
              std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
              std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
              const int domainPhysicalGroupDim,
              const int domainPhysicalGroupTag)
{
    // Only considers the domain of interrest.
    int physicalGroupDim = domainPhysicalGroupDim;
    int physicalGroupTag = domainPhysicalGroupTag;

    // Retrieves the node tags and coordinates.
    gmsh::model::mesh::getNodesForPhysicalGroup(physicalGroupDim, physicalGroupTag, domainNodeTags, domainNodeCoords);

    // Constructs the "domainNodeIndices" map associating the index of each node tag in the "domainNodeCoords" vector.
    for(std::size_t index = 0; index < domainNodeTags.size(); index++)
        domainNodeIndices[domainNodeTags[index]] = index;

    // Loops over the entities of the domain to fill the "elementTypes" and "elementTags" vectors and the "domainEntityIndices"
    // and "domainElementNodeTags" maps.
    gmsh::model::getEntitiesForPhysicalGroup(physicalGroupDim, physicalGroupTag, domainEntityTags);
    domainElementTypes.resize(domainEntityTags.size());
    domainElementTags.resize(domainEntityTags.size());
    for(std::size_t j = 0; j < domainEntityTags.size(); j++)
    {        
        // Retrieves every 'element related' data from the tag of its entity.
        std::vector<int> elementTypes;
        std::vector<std::vector<std::size_t>> elementTags;
        std::vector<std::vector<std::size_t>> nodeTags;
        gmsh::model::mesh::getElements(elementTypes, elementTags, nodeTags, physicalGroupDim, domainEntityTags[j]);

        domainElementTypes[j].resize(elementTypes.size());
        domainElementTags[j].resize(elementTypes.size());
        for(std::size_t k = 0; k < elementTypes.size(); k++)
        {
            // Initializes a vector that will contain the tags of the nodes of the current element (the number of nodes only
            // depends on the element type).
            int nbNodesPerElement = nodeTags[k].size() / elementTags[k].size();
            std::vector<std::size_t> elementNodeTags(nbNodesPerElement);

            domainElementTypes[j][k] = elementTypes[k];
            domainElementTags[j][k].resize(elementTags[k].size());

            for(std::size_t l = 0; l < elementTags[k].size(); l++)
            {
                domainElementTags[j][k][l] = elementTags[k][l];
                
                // Find the proper node tags of the elements.
                int n = 0;
                for(std::size_t m = l*nbNodesPerElement; m < (l+1)*nbNodesPerElement; m++)
                {
                    elementNodeTags[n] = nodeTags[k][m];
                    n++;
                }
                
                // Stores it in the "domainElementNodeTags" map (element tag -> vector of node tags).
                domainElementNodeTags[elementTags[k][l]] = elementNodeTags;
            }
        }
    }
}

// From the "elementVector" vector containing the information of all elements, constructs the "boundaryNodeToElements" map
// associating to each boundary node tag, a vector of element tags of the contiguous elements. The vector of element tags
// contains 2 element excepts in case of a mid element node (for higher order meshes), where only one element tag is stored.
void computeNodeToElements(std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                           const std::vector<elementStruct> &elementVector)
{
    // Loops over the elements and add the element tag to each of its corresponding nodes.
    for(std::size_t i = 0; i < elementVector.size(); i++)
    {
        if(!boundaryNodeToElements.count(elementVector[i].nodeTags[0]))
            boundaryNodeToElements[elementVector[i].nodeTags[0]] = {i};
        else
            boundaryNodeToElements[elementVector[i].nodeTags[0]].push_back(i);

        if(!boundaryNodeToElements.count(elementVector[i].nodeTags[1]))
            boundaryNodeToElements[elementVector[i].nodeTags[1]] = {i};
        else
            boundaryNodeToElements[elementVector[i].nodeTags[1]].push_back(i);

        if(elementVector[i].midNodeTag != 0)
            boundaryNodeToElements[elementVector[i].midNodeTag] = {i};
    }
}

// Computes and stores the potential of each interior node in the "internalNodalPhi" (it can be accessed via the tag of the
// node). "domainNodeTags" and "domainNodeCoords" are vectors containing all the node tags and coordinates (respectivelly) of the
// domain. "domainNodeIndices" is a map associating the index of each node tag for accessing the "domainNodeCoords" vector.
// "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM
// domain. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or numerically
// ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
// "boundaryNodeToElements" is a map associating to each boundary node tag, a vector of element tags of the contiguous elements.
void computeInternalPhi(std::map<int, double> &internalNodalPhi,
                        const std::vector<double> &domainNodeCoords,
                        const std::vector<std::size_t> &domainNodeTags,
                        std::map<std::size_t, std::size_t> &domainNodeIndices,
                        const std::vector<elementStruct> &elementVector,
                        const std::string integrationType,
                        std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                        const bool &solutionType)
{
    double x = 0;
    double y = 0;
    double phi = 0;
    double h = 0;
    double g = 0;
    std::vector<double> internalNodalPhiTmp(domainNodeTags.size(), 0); // Used for parallelization.

    // Loops over all the internal nodes coordinates, computes the potential at that point and stores it in
    // "internalNodalPhiTmp".
    #pragma omp parallel for schedule(static) private(x, y, phi, h, g) shared(domainNodeTags, domainNodeCoords, elementVector, internalNodalPhi, boundaryNodeToElements, solutionType)
    for(std::size_t i = 0; i < domainNodeTags.size(); i++)
    {
        // Do not consider the boundary nodes.
        if(!boundaryNodeToElements.count(domainNodeTags[i]))
        {
            // Gets the coordinates of the node.
            x = domainNodeCoords[3*domainNodeIndices[domainNodeTags[i]]];
            y = domainNodeCoords[3*domainNodeIndices[domainNodeTags[i]] + 1];

            phi = 0;
            // Loop over all the boundary element in order to compute the potential.
            for(size_t j = 0; j < elementVector.size(); j++)
            {
                h = computeAnalyticalH(x, y, elementVector[j]);
                if(!solutionType)
                    g = computeNumericalG(x, y, elementVector[j], integrationType);
                else
                    g = computeAnalyticalG(x, y, elementVector[j]);

                phi += h*elementVector[j].bcValue[0] - g*elementVector[j].bcValue[1];
            }
            internalNodalPhiTmp[i] = phi;
        }
    }

    // Finally writes the potential in the "internalNodePhi" map.
    for(std::size_t i = 0; i < domainNodeTags.size(); i++)
        internalNodalPhi[domainNodeTags[i]] = internalNodalPhiTmp[i];
}


// Binds the potential at each node to the right elements. "phi" is a data structure storing the potiential of each node (fourth
// dimension) of each element tag (third dimension) of each element type (second dimension) of each entity tag (first dimension).
// "domainElementTags" stores the tag of each element (third dimension) of each element type (second dimension) of each entity
// tag (first dimension). "domainElementTypes" stores the type of each element (second dimension) of each entity (first
// dimension). The "domainElementNodeTags" map associates to each element tag a vector of node tags representing the nodes of
// that element. "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of
// the current BEM domain. "internalNodalPhi" is a map associating the value of the potential to each node tag of the domain
// (boundaries excuded). "domainPhysicalGroupTag" is the tag of the physical domain of interrest. "boundaryNodeToElements" is a
// map associating to each boundary node tag, a vector of element tags of the contiguous elements.
void elementNodeData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                     const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                     const std::vector<std::vector<int>> &domainElementTypes,
                     std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                     const std::vector<elementStruct> &elementVector,
                     std::map<int, double> &internalNodalPhi,
                     const int domainPhysicalGroupTag,
                     std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements)
{       
    std::map<std::size_t, std::vector<std::size_t>> position;
    phiInitialization(phi, position, domainElementTypes, domainElementTags);

    std::vector<size_t> domainBoundaryElementTags;
    boundaryElementTags(domainBoundaryElementTags, domainElementTypes, domainElementTags, elementVector, domainPhysicalGroupTag);

    // Loops over all the element that have node(s) in common with the boundary, and computes (or decides) what is the value of
    // the potential at these boundary nodes.
    #pragma omp parallel for schedule(guided)
    for(std::size_t i = 0; i < domainBoundaryElementTags.size(); i++)
    {
        // Access everything related to the current element.
        std::size_t elementTag = domainBoundaryElementTags[i];
        int entityTagIndex = position[elementTag][0];
        int elementTypeIndex = position[elementTag][1];
        int elementTagIndex = position[elementTag][2];
        int nbNodes = phi[entityTagIndex][elementTypeIndex][elementTagIndex].size();
        std::vector<std::size_t> nodeTags = domainElementNodeTags[elementTag];
        
        // Loops on the nodes of each boundary element.
        for(int j = 0; j < nbNodes; j++)
        {
            std::size_t elementNodeTag = nodeTags[j];   // Tag of each node.

            // Only interrested in boundary nodes.
            if(boundaryNodeToElements.count(elementNodeTag))
            {
                // Retrieves the tags of the contiguous boundary elements of the node.
                std::vector<std::size_t> commonBoundaryElementIndices = boundaryNodeToElements[elementNodeTag];

                // Mid node of a boundary element => take the value of the element.
                if(commonBoundaryElementIndices.size() == 1)
                    phi[entityTagIndex][elementTypeIndex][elementTagIndex][j] = elementVector[commonBoundaryElementIndices[0]].bcValue[0];
                // Several cases that lead to either select the value of the potential on one element or computes the average of
                // the potential of two adjacent elements.
                else
                {
                    double currentPhi = 0.0;
                    int count = 0;

                    // Check the first contiguous element.
                    int adjacentElementIndex = commonBoundaryElementIndices[0];
                    std::size_t adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[0];
                    if(adjacentNodeTag == elementNodeTag)
                        adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[1];

                    // if "adjacentNodeTag" is in "nodeTags"
                    if(std::count(nodeTags.begin(), nodeTags.end(), adjacentNodeTag))
                        count++;
                    
                    currentPhi += elementVector[adjacentElementIndex].bcValue[0];

                    // Check the second contiguous element.
                    adjacentElementIndex = commonBoundaryElementIndices[1];
                    adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[0];
                    if(adjacentNodeTag == elementNodeTag)
                        adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[1];

                    // if "adjacentNodeTag" is in "nodeTags"
                    if(std::count(nodeTags.begin(), nodeTags.end(), adjacentNodeTag))
                    {
                        if(count == 0)
                            currentPhi = elementVector[adjacentElementIndex].bcValue[0];
                        else
                        {
                            currentPhi += elementVector[adjacentElementIndex].bcValue[0];
                            currentPhi /= 2;
                        }
                    }
                    else
                        if(count == 0)
                        {
                            currentPhi += elementVector[adjacentElementIndex].bcValue[0];
                            currentPhi /= 2;
                        }

                    phi[entityTagIndex][elementTypeIndex][elementTagIndex][j] = currentPhi;
                }
            }
        }
    }

    elementNodalPhiAssociation(phi, domainElementTags, domainElementNodeTags, internalNodalPhi);
}

// Initializes the "phi" data structure in order for it to have the proper dimensions. It is a data structure storing the
// potiential of each node (fourth dimension) of each element tag (third dimension) of each element type (second dimension) of
// each entity tag (first dimension). The "position" map associates to each of the element tag of the domain the 3 indices for
// accessing it in the "domainElementTags" or "phi" data structure. "domainElementTags" stores the tag of each element (third
// dimension) of each element type (second dimension) of each entity tag (first dimension). "domainElementTypes" stores the type
// of each element (second dimension) of each entity (first dimension).
void phiInitialization(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                       std::map<std::size_t, std::vector<std::size_t>> &position,
                       const std::vector<std::vector<int>> &domainElementTypes,
                       const std::vector<std::vector<std::vector<size_t>>> &domainElementTags)
{
    // First dimension: entity tags.
    phi.resize(domainElementTypes.size());
    for(std::size_t i = 0; i < domainElementTypes.size(); i++)
    {
        // Second dimension: element types.
        phi[i].resize(domainElementTypes[i].size());
        
        #pragma omp parallel for schedule(guided) shared(phi, domainElementTypes, domainElementTags, position)
        for(std::size_t j = 0; j < domainElementTypes[i].size(); j++)
        {
            // "getElementProperties" is called in order to access the number of nodes of each element (only depends on the
            // element type).
            std::string elementName;
            int dim;
            int order;
            int nbNodes;
            std::vector<double> localNodeCoord;
            int nbPrimaryNodes;
            gmsh::model::mesh::getElementProperties(domainElementTypes[i][j], elementName, dim, order, nbNodes, localNodeCoord,
                                                    nbPrimaryNodes);

            // Third dimension: element tags.
            phi[i][j].resize(domainElementTags[i][j].size());
            for(std::size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                // Fourth dimension: node tags.
                phi[i][j][k].resize(nbNodes, HUGE_VAL); // First set to a high value
                std::vector<std::size_t> pos = {i, j, k};

                #pragma omp critical
                    position[domainElementTags[i][j][k]] = pos; // Stores the indices
            }
        }
    }
}

// Fills the "domainBoundaryElementTags" vector containing the tag of every elements possessing at least one node in common with
// the boundary. "domainElementTags" stores the tag of each element (third dimension) of each element type (second dimension) of
// each entity tag (first dimension). "domainElementTypes" stores the type of each element (second dimension) of each entity
// (first dimension). "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file)
// of the current BEM domain. "domainPhysicalGroupTag" is the tag of the physical domain of interrest.
void boundaryElementTags(std::vector<std::size_t> &domainBoundaryElementTags,
                         const std::vector<std::vector<int>> &domainElementTypes,
                         const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                         const std::vector<elementStruct> &elementVector,
                         const int domainPhysicalGroupTag)
{
    // #pragma omp parallel
    {
        std::vector<std::size_t> localElementTags;  // For parallelization.

        // Loops over each boundary element and retrieves the element in contact with its boundary nodes.
        // #pragma omp for schedule(guided)
        for(std::size_t i = 0; i < elementVector.size(); i++)
        {
            int elementType;
            std::vector<size_t> nodeTags;
            int entityDim;
            int entityTag;
            std::vector<int> physicalTags;

            // Adds the elements tags to the local vector. 
            for(std::size_t j = 0; j < elementTags.size(); j++)
            {
                // Retrieves the physical group to which the element belongs because it is possible that the element belongs to
                // an adjacent physical surface.
                int elementType;
                std::vector<size_t> nodeTags;
                int entityDim;
                int entityTag;
                std::vector<int> physicalTags;
                // #pragma omp critical
                {
                    gmsh::model::mesh::getElement(elementTags[j], elementType, nodeTags, entityDim, entityTag);
                    gmsh::model::getPhysicalGroupsForEntity(entityDim, entityTag, physicalTags);
                }
                
                // Add the element tag (to a local 'thread safe' vector) if the physical tag is the right one.
                for(std::size_t k = 0; k < physicalTags.size(); k++)
                    if(physicalTags[k] == domainPhysicalGroupTag)
                        localElementTags.push_back(elementTags[j]);
            }
        }

        // // Fills the "domainBoundaryElementTags" vector.
        // #pragma omp critical
        domainBoundaryElementTags.insert(domainBoundaryElementTags.end(), localElementTags.begin(), localElementTags.end());
    }
    

    // Allows duplicate tags to be removed. If an element has several nodes in common with the boundary, it has been counted
    // several times before.
    sort( domainBoundaryElementTags.begin(), domainBoundaryElementTags.end() );
    domainBoundaryElementTags.erase( unique( domainBoundaryElementTags.begin(), domainBoundaryElementTags.end() ),
                                     domainBoundaryElementTags.end() );
}

// Associates the potential to the nodes of each elements (if not already done) in the "phi" data structure. It is a data
// structure storing the potiential of each node (fourth dimension) of each element tag (third dimension) of each element type
// (second dimension) of each entity tag (first dimension). "domainElementTags" stores the tag of each element (third dimension)
// of each element type (second dimension) of each entity tag (first dimension). The "domainElementNodeTags" map associates to
// each element tag a vector of node tags representing the nodes of that element. "internalNodalPhi" is a map associating the
// value of the potential to each node tag of the domain (boundaries excuded).
void elementNodalPhiAssociation(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                                const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                                std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                                std::map<int, double> &internalNodalPhi)
{
    // Loops over each nodes of the domain (entity tags, then element types, than element tags, the node tags).
    for(std::size_t i = 0; i < domainElementTags.size(); i++)
    {
        for(std::size_t j = 0; j < domainElementTags[i].size(); j++)
        {
            #pragma omp parallel for schedule(guided)
            for(std::size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                std::vector<std::size_t> nodeTags = domainElementNodeTags[domainElementTags[i][j][k]];
                for(std::size_t l = 0; l < nodeTags.size(); l++)
                {
                    std::size_t elementNodeTag = nodeTags[l];
                    
                    // If the potential has not already been modified in "phi" and if the node does not belong to the boundary,
                    // retrieves the potential of that node thanks to the "interNodalPhi".
                    if(phi[i][j][k][l] == HUGE_VAL && internalNodalPhi.count(elementNodeTag))
                        phi[i][j][k][l] = internalNodalPhi[elementNodeTag];
                }
            }
        }
    }
}

// Computes the potential (if "computePhi" is set to 'true') and the electric field (if "computeElectricField" is set to 'true')
// and stores them in the "phi" and "electricField" datastructures respectivelly. These assign the value of the potential /
// electric field to each element tag (third dimension) of each element type (second dimension) of each entity tag (first
// dimension) of the domain. The fourth dimension is either of size 1 (for the potential) or 3 (for the electric field).
// "domainEntityTags" is a vector containing the tag of each entity of the domain. "domainElementTypes" stores the types of the
// elements of the the domain. The first dimension of this vector represents the different entities and the second, the different
// element types. "domainElementTags" stores the tags of the elements of the domain. The first dimension of this vector
// represents the different entities, the second, the different element types and the third, the different element tags.
// "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM
// domain. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or numerically
// ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
void computeElementData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::vector<std::vector<std::vector<std::vector<double>>>> &electricField,
                        const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<int> &domainEntityTags,
                        const std::vector<elementStruct> &elementVector,
                        const bool solutionType,
                        const std::string integrationType,
                        const bool computePhi,
                        const bool computeElectricField)
{
    std::vector<double> barycenters;

    // Loops over all the entity tags and element types.
    phi.resize(domainEntityTags.size());
    electricField.resize(domainEntityTags.size());
    for(std::size_t i = 0; i < domainEntityTags.size(); i++)
    {
        phi[i].resize(domainElementTypes[i].size());
        electricField[i].resize(domainElementTypes[i].size());
        for(std::size_t j = 0; j < domainElementTypes[i].size(); j++)
        {
            phi[i][j].resize(domainElementTags[i][j].size());
            electricField[i][j].resize(domainElementTags[i][j].size());
            
            // Retrieves the barycenters of each elements.
            gmsh::model::mesh::getBarycenters(domainElementTypes[i][j], domainEntityTags[i], false, false, barycenters);
            
            // Loops over each barycenter and compute the potential and electric field at that point.
            #pragma omp parallel for schedule(static)
            for(size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                double x = barycenters[3*k];
                double y = barycenters[3*k + 1];

                double currentPhi = 0;
                double gradPhiX = 0;
                double gradPhiY = 0;

                double localCurrentPhi = 0;
                double localGradPhiX = 0;
                double localGradPhiY = 0;

                double h, hX, hY;
                double g, gX, gY;
                
                // Loops over all the boundary elements in order to compute the potential and the electric field.
                for(std::size_t el = 0; el < elementVector.size(); el++)
                {
                    if(computePhi)
                    {
                        h = computeAnalyticalH(x, y, elementVector[el]);
                        if(!solutionType)
                            g = computeNumericalG(x, y, elementVector[el], integrationType);
                        else
                            g = computeAnalyticalG(x, y, elementVector[el]);

                        localCurrentPhi += h*elementVector[el].bcValue[0] - g*elementVector[el].bcValue[1];
                    }
                    if(computeElectricField)
                    {
                        if(!solutionType)
                        {
                            hX = computeNumericalGradHx(x, y, elementVector[el], integrationType);
                            gX = computeNumericalGradGx(x, y, elementVector[el], integrationType);
                            hY = computeNumericalGradHy(x, y, elementVector[el], integrationType);
                            gY = computeNumericalGradGy(x, y, elementVector[el], integrationType);
                        }
                        else
                        {
                            hX = computeAnalyticalGradHx(x, y, elementVector[el]);
                            gX = computeAnalyticalGradGx(x, y, elementVector[el]);
                            hY = computeAnalyticalGradHy(x, y, elementVector[el]);
                            gY = computeAnalyticalGradGy(x, y, elementVector[el]);
                        }

                        localGradPhiX += hX*elementVector[el].bcValue[0] - gX*elementVector[el].bcValue[1];
                        localGradPhiY += hY*elementVector[el].bcValue[0] - gY*elementVector[el].bcValue[1];
                    }
                }

                // Updates the potential and electric field (for the parallelization)
                currentPhi += localCurrentPhi;
                gradPhiX += localGradPhiX;
                gradPhiY += localGradPhiY;

                // Writes the data in the data structures.
                phi[i][j][k].resize(1, currentPhi);
                electricField[i][j][k].resize(3);
                electricField[i][j][k][0] = -gradPhiX;
                electricField[i][j][k][1] = -gradPhiY;
                electricField[i][j][k][2] = 0.0;
            }
        }
    }
}

// Deals with the display in gmsh. "viewTag" is the view tag of the data. "modelName" is the name of the model. "fileName" is the
// name of the output file. "dataType" is either "ElementData" or "ElementNodeData" (which are the supported type of this
// application). "domainElementTags" stores the tags of the elements of the domain. The first dimension of this vector represents
// the different entities, the second, the different element types and the third, the different element tags. "data" is a data
// structure storing the data to be displayed. The first three dimensions follow the same structure as the one of the
// "elementNodeTags" and the fourth one is contains what needs to be displayed. "viewIndex" if the index of the view.
void displayData(const int viewTag,
                 const std::string modelName,
                 const std::string fileName,
                 const std::string dataType,
                 const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                 const std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                 const int viewIndex)
{
    for(std::size_t i = 0; i < data.size(); i++)
        for(std::size_t j = 0; j < data[i].size(); j++)
            gmsh::view::addModelData(viewTag, 0 , modelName, dataType, domainElementTags[i][j], data[i][j]);
    
    //gmsh::view::write(viewTag, fileName);

    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].IntervalsType", 3);
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].MaxRecursionLevel", 3);
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].TargetError", -0.0001);
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].Visible", 0); // preventing the superimposed plotting
}