#include "functionsBEM.hpp"

// Fills the "electrostaticPressure" map that gives the electrostatic pressure associated with each (boundary) element tag.
// "nbViews" is the counter of views (useful if other views have to be generated than with this solver). "nodalDisplacements" is
// a map associating the (x,y) displacement to a serie of node tags (if the position of the nodes are not striclty the one
// extracted from the .geo file). "postProcessing" indicates whether the post-processing should be done ('true') or not
// ('false'). "iteration" represents the number of times the solver has been called (some information is printed on the first
// iteration). "untangleMesh" unstangles every BEM domain if set to 'true' (in case the "nodalDisplacements" map contains
// non-zero displacements).
void solverBEM(std::map<int, double> &electrostaticPressure,
               int & nbViews,
               std::map<int,std::pair<double,double>> & nodalDisplacements,
               const bool postProcessing,
               const int iteration,
               const bool untangleMesh)
{
    bool computePhi = true;             // true: computes the potential.
    bool computeElectricField = true;   // true: computes the electric field.

    // Retrieves the names of the BEM domains as well as their physical dim/tag.
    gmsh::vectorpair physicalGroupsDimTags;
    gmsh::model::getPhysicalGroups(physicalGroupsDimTags);

    std::vector<std::string> domainNames;
    std::vector<std::pair<int, int>> domainPhysicalGroupDimTags;

    for(std::size_t i = 0; i < physicalGroupsDimTags.size(); i++)
    {
        int physicalGroupsDim = physicalGroupsDimTags[i].first;
        int physicalGroupsTag = physicalGroupsDimTags[i].second;
        std::string name;
        gmsh::model::getPhysicalName(physicalGroupsDim, physicalGroupsTag, name);

        std::stringstream ss(name);
        std::string word;
        std::vector<std::string> words;
        while(getline(ss, word, '_'))
            words.push_back(word);
        
        if(words.size() == 3 && words[0].compare("BEM") == 0 && words[1].compare("domain") == 0)
        {
            domainNames.push_back(words[2]);
            domainPhysicalGroupDimTags.push_back({physicalGroupsDim, physicalGroupsTag});
        }
    }

    // Useful for dealing with the untangler.
    if(untangleMesh)
        // Untangles each BEM domain.
        for(std::size_t i = 0; i < domainNames.size(); i++)
        {
            int dim = domainPhysicalGroupDimTags[i].first;
            int tag = domainPhysicalGroupDimTags[i].second;

            std::vector<int> tags;
            gmsh::model::getEntitiesForPhysicalGroup(dim, tag, tags);

            gmsh::vectorpair dimTagsBis(tags.size());
            for(size_t i = 0; i < tags.size(); i++)
                dimTagsBis[i] = std::pair<int,int>(dim, tags[i]);

            gmsh::model::mesh::optimize("UntangleMeshGeometry", false, 1, dimTagsBis);

        }


    if (domainNames.size() == 0)
    {
        std::cerr << "Error: No BEM domain found ! Usage: <BEM_domain_'domainName'>\n";
        exit(0);
    }

    if(iteration == 1)
    {
        std::cout << "--------------------[BEM SOLVER]--------------------\n";
        std::cout << "The model contains " << domainNames.size() << " BEM domain(s):\n";
    }

    // Generates viewTags only if the postProcessing is enabled.
    int phiViewTag = 0;
    int electricFieldViewTag = 0;
    if(postProcessing)
    {
        if(computePhi)
            phiViewTag = gmsh::view::add("Phi");
        if(computeElectricField)
            electricFieldViewTag = gmsh::view::add("Electric Field");
    }

    // Runs the solver for each BEM domain.
    for(std::size_t i = 0; i < domainNames.size(); i++)
        singleDomainSolverBEM(electrostaticPressure, nbViews, nodalDisplacements, postProcessing, iteration,
                              "BEM_domain_" + domainNames[i], phiViewTag, electricFieldViewTag, computePhi,
                              computeElectricField);

    if(iteration == 1)
        std::cout << "--------------------[BEM SOLVER]--------------------\n\n";

    // Updates the number of views.
    if(postProcessing)
    {
        if(computePhi)
            nbViews++;
        if(computeElectricField)
            nbViews++;
    }
}


// Fills the "electrostaticPressure" map. The "nodalDisplacements", "postProcessing" and "iteration" arguments are described in
// the description of the function "solverBEM". "domainName" is used for extracting everything in the appropriate independant BEM
// domain, "phiViewTag" and "electricFieldViewTag" are the view tag of the potential and of the electric field respectively.
// "computePhi" and "computeElectricField" indicates whether the potential and/or the electric field should be computed ('true')
// or not ('false').
void singleDomainSolverBEM(std::map<int, double> &electrostaticPressure,
                           const int nbViews,
                           std::map<int,std::pair<double,double>> & nodalDisplacements,
                           const bool postProcessing,
                           const int iteration,
                           const std::string domainName,
                           const int phiViewTag,
                           const int electricFieldViewTag,
                           const bool computePhi,
                           const bool computeElectricField)
{
    // Parameters:
    bool showTime = false;      // true: displays the different time measurements.

    bool convergence = false;   // true: computes the error between the analytical solution and the numerical one on a simple
                                // configuration (/!\ use this with the BEM_convergence.geo model).
    bool visualisation = true;  // true: 'ElementNodeData' visualization of the potential, false: 'ElementData' visualization of
                                // the potential (Remark: electric field is always displayed with 'ElementData' type).
    bool solutionType = true;   // true: analytical computations, false: numerical computations.

    // Used for displaying the time of different sections.
    // double start, end;
    double startTotal, endTotal;
    double start, end;
    startTotal = omp_get_wtime();

    // Initializes a vector of elements (that will contain all the elements of interrest)
    std::vector<elementStruct> elementVector;
    std::string integrationType = "Gauss10"; // Used for the numerical integrations when "solutionType" = 'false'.

    // Fills the vector of boundary elements.
    start = omp_get_wtime();

    int domainPhysicalGroupDim;
    int domainPhysicalGroupTag;
    std::size_t nbDirichletElements = fillElementVector(elementVector, domainPhysicalGroupDim, domainPhysicalGroupTag,
                                                        nodalDisplacements, domainName);

    end = omp_get_wtime();
    displayTime(start, end, "'fillElementVector'", showTime);

    if(iteration == 1)
        std::cout << "\n- BEM domain '" << domainName << "': contains " << elementVector.size() << " boundary elements.\n";

    // Constructs the linear system
    start = omp_get_wtime();
    
    // The matrix type is perhaps not the most efficient.
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(elementVector.size(), elementVector.size());
    Eigen::MatrixXd B = Eigen::MatrixXd::Zero(elementVector.size(), elementVector.size());
    Eigen::MatrixXd c = Eigen::MatrixXd::Zero(elementVector.size(), 1);
    fillSystem(A, B, c, elementVector, nbDirichletElements, integrationType, solutionType);

    end = omp_get_wtime();
    displayTime(start, end, "'fillSystem'", showTime);

    // Solves the linear system.
    start = omp_get_wtime();

    Eigen::MatrixXd b = B*c;
    Eigen::MatrixXd x = A.partialPivLu().solve(b);

    end = omp_get_wtime();
    displayTime(start, end, "Solve linear system", showTime);

    // Computes the electrostatic pressure on each boundary element.
    start = omp_get_wtime();
    computeElectrostaticPressure(electrostaticPressure, elementVector, x, nbDirichletElements, domainName);
    end = omp_get_wtime();
    displayTime(start, end, "'computeElectrostaticPressure'", showTime);

    // Visualization
    if(postProcessing)
    {
        int viewIndex = nbViews;
        
        // Retrieves everything related to domain of interrest.
        start = omp_get_wtime();

        std::vector<std::size_t> domainNodeTags;
        std::map<std::size_t, std::size_t> domainNodeIndices;
        std::vector<double> domainNodeCoords;

        std::vector<int> domainEntityTags;
        std::vector<std::vector<int>> domainElementTypes;
        std::vector<std::vector<std::vector<size_t>>> domainElementTags;
        std::map<std::size_t, std::vector<std::size_t>> domainElementNodeTags;

        getModel(domainNodeTags, domainNodeIndices, domainNodeCoords, domainEntityTags, domainElementTypes, domainElementTags,
                 domainElementNodeTags, domainPhysicalGroupDim, domainPhysicalGroupTag);

        end = omp_get_wtime();
        displayTime(start, end, "'getModel'", showTime);
        
        // The first index indicates the "entityTag", the second the "elementType", the third the "elementTag", and the fourth,
        // either the nodesTags associated with the element (for elementNodalPhi) or the computed values (for "elemental...") (1
        // for the potential and 3 for the electric field).
        std::vector<std::vector<std::vector<std::vector<double>>>> elementalPhi;
        std::vector<std::vector<std::vector<std::vector<double>>>> elementNodalPhi;
        std::vector<std::vector<std::vector<std::vector<double>>>> elementalElectricField;

        std::vector<std::string> names;
        gmsh::model::list(names);

        // Visualisation by ElementNodeData
        if(visualisation)
        {
            // Computes a map that stores 'nodeTag' (of the boundary) -> vector of adjacent 'elementIndices' (because only 1
            // element for higher order elements).
            std::map<std::size_t, std::vector<std::size_t>> boundaryNodeToElements;
            computeNodeToElements(boundaryNodeToElements, elementVector);

            // Computes the value of the potential at each node strictly included in the domain and stores it in the
            // "internalNodalPhi" map.
            start = omp_get_wtime();

            std::map<int, double> internalNodalPhi;
            computeInternalPhi(internalNodalPhi, domainNodeCoords, domainNodeTags, domainNodeIndices, elementVector,
                               integrationType, boundaryNodeToElements, solutionType);
            
            end = omp_get_wtime();
            displayTime(start, end, "'computeInternalPhi'", showTime);

            // Associates the value of the potential to each node of an element.
            start = omp_get_wtime();
            elementNodeData(elementNodalPhi, domainElementTags, domainElementTypes, domainElementNodeTags, elementVector,
                            internalNodalPhi, domainPhysicalGroupTag, boundaryNodeToElements);
            end = omp_get_wtime();
            displayTime(start, end, "'elementNodeData'", showTime);
            
            // Computes the electric field (with "ElementData" type).
            start = omp_get_wtime();
            computeElementData(elementalPhi, elementalElectricField, domainElementTags, domainElementTypes, domainEntityTags,
                               elementVector, solutionType, integrationType, false, computeElectricField);
            end = omp_get_wtime();
            displayTime(start, end, "'computeElementData' (only the electric field)", showTime);

            // Displays what is needed.
            if(computePhi)
            {
                displayData(phiViewTag, names[0], "results.msh", "ElementNodeData", domainElementTags, elementNodalPhi,
                            viewIndex);
                viewIndex++;
            }
            if(computeElectricField)
                displayData(electricFieldViewTag, names[0], "results.msh", "ElementData", domainElementTags,
                            elementalElectricField, viewIndex);
        }
        // Visualisation by ElementData
        else
        {
            // Computes the potential and the electric field.
            start = omp_get_wtime();
            computeElementData(elementalPhi, elementalElectricField, domainElementTags, domainElementTypes, domainEntityTags,
                               elementVector, solutionType, integrationType, computePhi, computeElectricField);
            end = omp_get_wtime();
            displayTime(start, end, "'computeElementData'", showTime);    

            // Displays what is needed.
            if(computePhi)
            {
                displayData(phiViewTag, names[0], "results.msh", "ElementData", domainElementTags, elementalPhi, viewIndex);
                viewIndex++;
            }
            if(computeElectricField)
                displayData(electricFieldViewTag, names[0], "results.msh", "ElementData", domainElementTags,
                            elementalElectricField, viewIndex);

            // Study of the error on a precise model.
            if(convergence)
            {   
                std::cout << "\tCONVERGENCE:\n";
                std::string convergenceModelName = "BEM_convergence";
                if(names[0].compare(convergenceModelName) == 0)
                    convergenceFun(domainEntityTags, domainElementTypes, domainElementTags, elementalPhi,
                                   elementalElectricField);
                else
                {
                    std::cout << "\tWarning: Cannot check convergence on model '" << names[0] << ".geo'. Use the '";
                    std::cout << convergenceModelName + ".geo" << "' model if you want to check for convergence.\n\n";
                }
            }
        }
    }

    endTotal = omp_get_wtime();
    displayTime(startTotal, endTotal, "total time of the code BEM", showTime);
}